const stripAnsi = require("strip-ansi")
const smartwrap = require("smartwrap")
const wcwidth = require("wcwidth")


const addPadding = (config, width) => {
  return width + config.paddingLeft + config.paddingRight
}


/**
 * Returns the widest cell give a collection of rows
 *
 * @param object columnOptions
 * @param array rows
 * @param integer columnIndex
 * @returns string
 */
const getMaxLength = (columnOptions, rows, columnIndex) => {

  let iterable

  // add a row that contains the header value, so we use that width too
  if(typeof columnOptions === "object" && columnOptions.value) {
    iterable = rows.slice()
    let z = new Array(iterable[0].length) // create a new empty row
    z[columnIndex] = columnOptions.value.toString()
    iterable.push(z)
  } else {
    // no header value, just use rows to derive max width
    iterable = rows
  }

  // iterable.forEach( row => {
  //   if(row[columnIndex] && stripAnsi(row[columnIndex].toString()).length > widest) {
  //     widest = wcwidth(row[columnIndex].toString())
  //   }
  // })

  let widest = iterable.reduce((prev, row) => {
    if(row[columnIndex]) {
      let width = wcwidth(stripAnsi(row[columnIndex].toString()))
      return ( width > prev) ? width : prev
    }
    return prev
  }, 0)

  return widest
}


/**
 * Get total width available to this table instance
 *
 *
 */
const getAvailableWidth = config => {
  if (process && ((process.stdout && process.stdout.columns) || (process.env && process.env.COLUMNS))) {
    // forked calls that do not inherit process.stdout must use process.env
    let viewport = (process.stdout && process.stdout.columns) ? process.stdout.columns : process.env.COLUMNS
    viewport = viewport - config.marginLeft

    // table width fixed
    if (config.width !== "auto" && /^\d{1,2}$/.test(config.width)) return config.width

    // table width percentage of (viewport less margin)
    if (config.width !== "auto" && /^\d{1,2}%$/.test(config.width)) {
      return (config.width.slice(0, -1) * .01) * viewport
    }

    // table width equals viewport less margin
    return viewport
  }

  // browser
  if (typeof window !== "undefined") return window.innerWidth // eslint-disable-line

  throw new Error ("Cannot auto discover available table width. Set table `width` option or export the environment variable COLUMNS")
}


module.exports.getStringLength = string => {
  // stripAnsi(string.replace(/[^\x00-\xff]/g,'XX')).length
  return wcwidth(stripAnsi(string))
}


module.exports.wrapCellText = (
  config,
  cellValue,
  columnIndex,
  cellOptions,
  rowType
) => {

  // ANSI chararacters that demarcate the start/end of a line
  const startAnsiRegexp = /^(\033\[[0-9;]*m)+/
  const endAnsiRegexp = /(\033\[[0-9;]*m)+$/

  // coerce cell value to string
  let string = cellValue.toString()

  // store matching ANSI characters
  let startMatches = string.match(startAnsiRegexp) || [""]

  // remove ANSI start-of-line chars
  string = string.replace(startAnsiRegexp, "")

  // store matching ANSI characters so can be later re-attached
  let endMatches = string.match(endAnsiRegexp) || [""]

  // remove ANSI end-of-line chars
  string = string.replace(endAnsiRegexp, "")

  let alignTgt

  switch(rowType) {
    case("header"):
      alignTgt = "headerAlign"
      break
    case("body"):
      alignTgt = "align"
      break
    default:
      alignTgt = "footerAlign"
  }

  // equalize padding for centered lines
  if(cellOptions[alignTgt] === "center") {
    cellOptions.paddingLeft = cellOptions.paddingRight = Math.max(
      cellOptions.paddingRight,
      cellOptions.paddingLeft,
      0
    )
  }

  const columnWidth = config.table.columnWidths[columnIndex]

  // innerWidth is the width available for text within the cell
  const innerWidth = columnWidth
    - cellOptions.paddingLeft
    - cellOptions.paddingRight
    - config.GUTTER

  if(typeof config.truncate === "string") {
    string = exports.truncate(string, cellOptions, innerWidth)
  } else {
    string = exports.wrap(string, cellOptions, innerWidth)
  }

  // format each line
  let strArr = string.split("\n").map( line => {
    line = line.trim()

    const lineLength = exports.getStringLength(line)

    // alignment
    if(lineLength < columnWidth) {
      let emptySpace = columnWidth - lineLength

      switch(true) {
        case(cellOptions[alignTgt] === "center"):
          emptySpace --
          let padBoth = Math.floor(emptySpace / 2),
            padRemainder = emptySpace % 2
          line = Array(padBoth + 1).join(" ")
            + line
            + Array(padBoth + 1 + padRemainder).join(" ")
          break

        case(cellOptions[alignTgt] === "right"):
          line = Array(emptySpace - cellOptions.paddingRight).join(" ")
            + line
            + Array(cellOptions.paddingRight + 1).join(" ")
          break

        default:
          line = Array(cellOptions.paddingLeft + 1).join(" ")
            + line
            + Array(emptySpace - cellOptions.paddingLeft).join(" ")
      }
    }

    // put ANSI color codes BACK on the beginning and end of string
    return startMatches[0] + line + endMatches[0]
  })

  return {
    output: strArr,
    width: innerWidth
  }
}


module.exports.truncate = (string, cellOptions, maxWidth) => {
  const stringWidth = wcwidth(string)

  if(maxWidth < stringWidth) {
    // @TODO give user option to decide if they want to break words on wrapping
    string = smartwrap(string, {
      width: maxWidth - cellOptions.truncate.length,
      breakword: true
    }).split("\n")[0]
    string = string + cellOptions.truncate
  }

  return string
}


module.exports.wrap = (string, cellOptions, innerWidth) => {
  let outstring = smartwrap(string, {
    errorChar: cellOptions.defaultErrorValue,
    minWidth: 1,
    trim: true,
    width: innerWidth
  })

  return outstring
}


module.exports.getColumnWidths = (config, rows) => {

  const availableWidth = getAvailableWidth(config)

  // array of spaces that cant be proportionately reduced for each row
  const nonResizableRowSpaces = []

  // iterate over the header if we have it, iterate over the first row
  // if we do not (to step through the correct number of columns)
  let iterable = (config.table.header[0] && config.table.header[0].length > 0)
    ? config.table.header[0] : rows[0]

  let widths = iterable.map((column, columnIndex) => {
    let result
    nonResizableRowSpaces[columnIndex] = 0 // track non resizable space for row

    switch(true) {
      // column width is specified in column header
      case(typeof column === "object" && (/^\d{1,2}$/.test(column.width))):
        result = column.width
        break

      // column width is a percentage of table width specified in column header
      case(typeof column === "object" && (/^\d{1,2}%$/.test(column.width))):
        result = (column.width.slice(0, -1) * .01) * availableWidth
        result = addPadding(config, result)
        break

      // 'auto' sets column width to its longest value in the initial data set
      default:
        let columnOptions = (config.table.header[0][columnIndex])
          ? config.table.header[0][columnIndex] : {}
        let measurableRows = (rows.length) ? rows : config.table.header[0]

        result = getMaxLength(columnOptions, measurableRows, columnIndex)

        // add spaces for padding if not centered
        // @TODO test with if not centered conditional
        result = addPadding(config, result)
    }

    // add space for gutter
    result = result + config.GUTTER

    return result
  })

  // calculate sum of all column widths (including marginLeft)
  let totalWidth = widths.reduce((prev, current) => prev + current)

  // if sum of all widths exceeds viewport, resize proportionately to fit
  if(totalWidth > availableWidth) {
    let prop = availableWidth / totalWidth
    let relativeWidths
    let totalRelativeWidths

    prop = prop.toFixed(2) - 0.01 // this wont be exact fit, but keeps us safe

    // when prop < 0 column cant be resized and totalWidth must overflow viewport
    if (prop > 0) {
      relativeWidths = widths.map(value => Math.floor(prop * value))
      totalRelativeWidths = relativeWidths.reduce((prev, current) => prev + current)
      widths = (totalRelativeWidths < totalWidth) ? relativeWidths : widths
    }
  }

  return widths
}
